"""Track time either as decorator or explicit."""
import datetime as dt
import logging
import math
import time
import types
from functools import wraps
from typing import Optional


class TimeTrackingWrapper:
    r"""
    Wrapper implementation of TimeTracking class.

    Use this implementation easily as decorator for functions, classes and class methods. Implement a custom function
    and decorate it for automatic time measure.

    .. code-block:: python

        @TimeTrackingWrapper
        def sleeper():
            print("start")
            time.sleep(1)
            print("end")

        >>> sleeper()
        start
        end
        INFO: foo finished after 00:00:01 (hh:mm:ss)

    """

    def __init__(self, func):
        """Construct."""
        wraps(func)(self)

    def __call__(self, *args, **kwargs):
        """Start time tracking."""
        with TimeTracking(name=self.__wrapped__.__name__):
            return self.__wrapped__(*args, **kwargs)

    def __get__(self, instance, cls):
        """Create bound method object and supply self argument to the decorated method."""
        return types.MethodType(self, instance)


class TimeTracking(object):
    """
    Track time to measure execution time.

    Time tracking automatically starts on initialisation and ends by calling stop method. Duration can always be shown
    by printing the time tracking object or calling get_current_duration. It is possible to start and stop time tracking
    by hand like

    .. code-block:: python

        time = TimeTracking(start=True)  # start=True is default and not required to set
        do_something()
        time.stop(get_duration=True)

    A more comfortable way is to use TimeTracking in a with statement like:

    .. code-block:: python

        with TimeTracking():
            do_something()

    The only disadvantage of the latter implementation is, that the duration is logged but not returned.
    """

    def __init__(self, start=True, name="undefined job", logging_level=logging.INFO):
        """Construct time tracking and start if enabled."""
        self.start = None
        self.end = None
        self._name = name
        self._logging = {logging.INFO: logging.info, logging.DEBUG: logging.debug}.get(logging_level, logging.info)
        if start:
            self._start()

    def _start(self) -> None:
        """Start time tracking."""
        self.start = time.time()
        self.end = None

    def _end(self) -> None:
        """Stop time tracking."""
        self.end = time.time()

    def _duration(self) -> float:
        """Get duration in seconds."""
        if self.end:
            return self.end - self.start
        else:
            return time.time() - self.start

    def __repr__(self) -> str:
        """Display current passed time."""
        return f"{dt.timedelta(seconds=math.ceil(self._duration()))} (hh:mm:ss)"

    def run(self) -> None:
        """Start time tracking."""
        self._start()

    def stop(self, get_duration=False) -> Optional[float]:
        """
        Stop time tracking.

        Will raise an error if time tracking was already stopped.
        :param get_duration: return passed time if enabled.

        :return: duration if enabled or None
        """
        if self.end is None:
            self._end()
        else:
            msg = f"Time was already stopped {time.time() - self.end}s ago."
            raise AssertionError(msg)
        if get_duration:
            return self.duration()

    def duration(self) -> float:
        """Return duration in seconds."""
        return self._duration()

    def __enter__(self):
        """Context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Stop time tracking on exit and log info about passed time."""
        self.stop()
        self._logging(f"{self._name} finished after {self}")
